# Module that manages shelling out commands to ffmpeg, with functions returning paths to the final video.

from .printer import cprint
from .commands.stage import StageData, VideoSlice

import os
import subprocess
from pathlib import Path
from typing import List


class VideoFailure(Exception):
	def __init__(self, vid="?") -> None:
		self.vid = vid
		super().__init__(self.vid)

class FailedToSlice(VideoFailure):
	pass

class FailedToConcat(VideoFailure):
	pass

class FailedToCleanUp(VideoFailure):
	pass


def slice_video(TEMP_DIR: Path, LOG_LEVEL: str, vslice: VideoSlice, i: int=1, total: int=1) -> Path:
	tmpfile = TEMP_DIR / f"{vslice.video_id}={i}.mp4"
	cprint(f"#rSlicing stage part ({i}/{total}) `#fM{vslice.video_id}#r` #d({vslice.ss} - {vslice.to})#r")

	cmd = [ "ffmpeg", "-hide_banner", "-ss", vslice.ss ]

	if vslice.to != "EOF":
		cmd += ["-to", vslice.to]

	cmd += [
		"-i", vslice.filepath, "-c", "copy",
		str(tmpfile), "-y", "-stats", "-loglevel", LOG_LEVEL
	]
	
	result = subprocess.run(cmd)

	if result.returncode != 0:
		raise FailedToSlice(vslice.video_id)
	
	return tmpfile


def concat_video(TEMP_DIR: Path, LOG_LEVEL: str, stage_id: str, slice_paths: List[Path]) -> Path:
	# first create file list in temp dir
	list_path = TEMP_DIR / f"concat-{stage_id}.txt"
	with open(str(list_path), "w") as f:
		for path in slice_paths:
			f.write(f"file '{path.name}'\n")

	# then do subprocess for concat list
	concat_path = TEMP_DIR / f"concat-{stage_id}.mp4"
	cmd = [
		"ffmpeg", "-hide_banner", "-f", "concat",
		"-safe", "0", "-i", str(list_path),
		"-c", "copy", str(concat_path),
		"-y", "-stats", "-loglevel", LOG_LEVEL
	]
	
	print()
	cprint(f"#rConcatentating videos for `#fM{stage_id}#r`")
	
	# we need to hop directories real fast to perform this function.
	# we'll come back when we're done.
	cwd = os.getcwd()
	os.chdir(str(TEMP_DIR))

	result = subprocess.run(cmd)

	if result.returncode != 0:
		raise FailedToConcat()

	os.chdir(cwd)

	cprint(f"Cleaning up after stage `#fM{stage_id}#r`...")

	try:
		os.remove(str(list_path))
	except Exception as e:
		raise FailedToCleanUp(e)

	return concat_path


def process_stage(conf: dict, stage: StageData) -> Path:
	tempdir = Path(conf["temp_dir"])
	loglevel = conf["ffmpeg_loglevel"]

	# slice all the slices
	slices = len(stage.slices)
	slice_paths = [slice_video(tempdir, loglevel, stage.slices[x], x, slices) for x in range(slices)]

	# edge case of one video
	if len(slice_paths) == 1:
		return slice_paths[0]

	# concat all the slices
	concat_path = concat_video(tempdir, loglevel, stage.id, slice_paths)

	# clean up old slices
	for path in slice_paths:
		try:
			os.remove(path)
		except Exception as e:
			raise FailedToCleanUp(e)

	# return the path of the concated vid
	return concat_path



# Initialization of VodBot, setting up config, etc.

import json
import re
from os.path import isabs

import vodbot.util as util
from vodbot.printer import cprint

vodbotdir = util.vodbotdir
DEFAULT_CONF = util.DEFAULT_CONF
DEFAULT_CONF_PATH = util.DEFAULT_CONF_PATH


def create_config(args):
	channels = []
	timezone = "" # switch to UTC only
	
	# these are for determining directories.
	# we need to make sure any that are already provided are absolute right now.
	# we do it now so that the user doesn't enter in anything before receiving
	# an error that boots them out of the program.
	parts = [
		[args.voddir, "VODs", DEFAULT_CONF['vod_dir']],
		[args.clipdir, "Clips", DEFAULT_CONF['clip_dir']],
		[args.tempdir, "temporary data", DEFAULT_CONF['temp_dir']],
		[args.stagedir, "staged data", DEFAULT_CONF['stage_dir']]
	]

	for part in parts:
		if part[0]:
			if not isabs(part[0]):
				util.exit_prog(150, f"Path `{part[0]}` for {part[1]} is not an absolute path.")
			part.append(part[0])

	# Ask for channels
	if not args.channels:
		cprint("Enter the login names of the Twitch channels you'd like to have archived.")
		cprint("When you're done, just leave the input empty and press enter.")
		cprint("Example: the url is `https://twitch.tv/notquiteapex`, enter `notquiteapex`")
		while True:
			channel = input("> ")
			if channel == "" and len(channels) > 0:
				break
			elif channel == "" and len(channels) == 0:
				cprint("#fRError#r, no channels given.")
			elif not re.match(r"^[a-zA-Z0-9][\w]{0,24}$", channel): # https://discuss.dev.twitch.tv/t/3855/4
				cprint("#fRError#r, channel names only contain characters A-Z, 0-9, and '_'. They also can't start with '_'.")
			else:
				channels += [channel]
	else:
		channels = args.channels

	# Ask for timezone as UTC string
	# We only absolutely need the hours and minutes offset, and we regex it out of the string.
	if not args.timezone:
		cprint("Enter the UTC timezone for timedate referencing. Only use +, -, and numbers.")
		cprint("#dExample: `+0000` for UTC, `-0500` for America's EST. Default: `+0000`#r")
		while True:
			tz = input("> ")
			if tz == "":
				tz = "+0000"
				break
			elif not re.match(r"^[+-]\d\d\d\d$", tz):
				cprint("Error, UTC string not recognized.")
			else:
				break
	else:
		timezone = args.timezone
		# check that entered timezone is valid

	# ask for directories (any provided by the terminal arguments are already handeled.)
	cprint("Now let's get some directories to store data. The entered paths must be absolute, not relative.")
	cprint("If you'd like to use the default location listed, just leave the input blank and press enter.")

	for part in parts:
		if not part[0]:
			cprint(f"Enter where {part[1]} should be stored.\n#dDefault: `#l{part[2]}`#r")
			while True:
				inpdir = input("> ")
				if inpdir == "":
					part.append(part[2])
					break
				elif isabs(inpdir):
					# TODO: check if directory can be created? a file might already exist?
					part.append(inpdir)
					break
				else:
					cprint(f"#fRError#r, directory `#l{inpdir}#r` is not an absolute path for a directory.")

	# ready to write it all, go!
	cprint("#dWriting config...#r")
	# Edit default config variable.
	DEFAULT_CONF['twitch_channels'] = channels
	DEFAULT_CONF['stage_timezone'] = timezone
	DEFAULT_CONF['vod_dir'] = parts[0][3]
	DEFAULT_CONF['clip_dir'] = parts[1][3]
	DEFAULT_CONF['temp_dir'] = parts[2][3]
	DEFAULT_CONF['stage_dir'] = parts[3][3]
	# the file gets written outside this function


def run(args):
	# test write the config.
	try:
		util.make_dir(args.output.parent)
		with open(str(args.output), "w") as f:
			pass
	except FileNotFoundError:
		util.exit_prog(67, f"Cannot create file at \"{args.output}\".")
	
	# see which config we should make
	if args.default:
		# generate default config
		cprint("#dCreating default config...#r")
	else:
		# generate config from inputs
		create_config(args)
	
	# create directories now
	util.make_dir(DEFAULT_CONF['vod_dir'])
	util.make_dir(DEFAULT_CONF['clip_dir'])
	util.make_dir(DEFAULT_CONF['temp_dir'])
	util.make_dir(DEFAULT_CONF['stage_dir'])

	# now write the config
	try:
		#util.make_dir(args.output.parent) # redundant
		with open(str(args.output), "w") as f:
			json.dump(DEFAULT_CONF, f, indent=4)
	except FileNotFoundError:
		util.exit_prog(68, f"Cannot create file at \"{args.output}\" despite being able to before.")

	# list the location of the config and say what can be edited outside this command
	cprint(f"#fGFinished#r, the config can be edited at `#l{str(vodbotdir / 'conf.json')}#r`.")



# Module to make API calls to Twitch.tv

from typing import List
from .itd import gql

import json


# for escaping strings, taken from json.py
ENCODE_DICT = {
    '\b': '\\b',
    '\f': '\\f',
    '\n': '\\n',
    '\r': '\\r',
    '\t': '\\t',
}

DECODE_DICT = dict((v,k) for k,v in ENCODE_DICT.items())

# TODO: change to enum?
CHAT_STATE = [
	"PUBLISHED", "UNPUBLISHED",
	"PENDING_REVIEW", "PENDING_REVIEW_SPAM",
	"DELETED"
]

class Vod:
	def __init__(self,
		id:str, user_id:str, user_login:str, user_name:str, title:str,
		created_at:str, length:int, game_id:str="", game_name:str="",
		has_chat:bool = False
	):
		self.id = id
		self.user_id = user_id
		self.user_login = user_login
		self.user_name = user_name

		self.game_id = game_id
		self.game_name = game_name

		self.title = title
		self.created_at = created_at
		self.length = length
		
		self.url = f"twitch.tv/videos/{self.id}"

		self.has_chat = has_chat
	
	def __repr__(self):
		return f"VOD({self.id}, {self.created_at}, {self.user_name}, {self.created_at}, {self.duration})"
	
	def write_meta(self, filename):
		jsondict = {
			"id": self.id,
			"user_id": self.user_id,
			"user_login": self.user_login,
			"user_name": self.user_name,
			"game_id": self.game_id,
			"game_name": self.game_name,
			"title": self.title,
			"created_at": self.created_at,
			"length": self.length,
			"has_chat": self.has_chat
		}
		
		with open(filename, "w") as f:
			json.dump(jsondict, f, sort_keys=True, indent=4)


class Clip:
	def __init__(self, id:str, slug:str, title:str, created_at:str, 
		user_id:str, user_login:str, user_name:str,
		clipper_id:str, clipper_login:str, clipper_name:str,
		game_id:str, game_name:str, view_count:int, length:int,
		offset:int, video_id:str
	):
		self.id = id # id of the clip itself
		self.slug = slug # the url ending to identify the clip
		self.user_id = user_id # user streaming
		self.user_login = user_login
		self.user_name = user_name

		self.clipper_id = clipper_id # user who clipped stream
		self.clipper_login = clipper_login
		self.clipper_name = clipper_name

		self.game_id = game_id # the "game" being played
		self.game_name = game_name

		self.title = title # clip metadata
		self.created_at = created_at
		self.view_count = view_count
		self.length = length

		self.offset = offset # offset from the start of the stream the clip was made
		self.video_id = video_id # id of the video stream the clip was made of
		
		self.url = f"twitch.tv/{self.user_name}/clip/{self.id}"
	
	def __repr__(self):
		return f"Clip({self.title}, {self.created_at}, {self.user_name}, {self.clipper_name})"
	
	def write_meta(self, filename):
		jsondict = {
			"id": self.id,
			"slug": self.slug,
			"user_id": self.user_id,
			"user_login": self.user_login,
			"user_name": self.user_name,
			"clipper_id": self.clipper_id,
			"clipper_login": self.clipper_login,
			"clipper_name": self.clipper_name,
			"game_id": self.game_id,
			"game_name": self.game_name,
			"title": self.title,
			"created_at": self.created_at,
			"view_count": self.view_count,
			"length": self.length,
			"offset": self.offset,
			"video_id": self.video_id
		}
		
		with open(filename, "w") as f:
			json.dump(jsondict, f, sort_keys=True, indent=4)


class Channel:
	def __init__(self, id:str, login:str, display_name:str, created_at:str):
		self.id = id
		self.login = login
		self.display_name = display_name
		self.created_at = created_at
	
	def __repr__(self):
		return f"Channel({self.id}, {self.display_name})"


class ChatMessage:
	def __init__(self,
		user:str, color:str, offset:int, msg:str=None, state:str=None,
		enc_msg:str=None, enc_state:str=None
	):
		self.user = user
		self.color = color
		self.offset = offset

		if msg is not None:
			self.msg = msg
			self.enc_msg = self.encode_message(self.msg)
		elif enc_msg is not None:
			self.enc_msg = enc_msg
			self.msg = self.decode_message(self.enc_msg)
		else:
			raise Exception(f"Could not determine message from constructor for ChatMessage. (\"{msg}\") (\"{enc_msg}\")")

		if state is not None:
			self.state = state
			self.enc_state = self.encode_state(self.state)
		elif enc_state is not None:
			self.enc_state = enc_state
			self.state = self.decode_state(self.enc_state)
		else:
			raise Exception(f"Could not determine state from constructor for ChatMessage. (\"{state}\") (\"{enc_state}\")")
		
	def __repr__(self):
		return f"ChatMessage(ofst={self.offset};{self.user}: {self.msg})"
		
	def write_dict(self):
		return {
			"ofst":self.offset_secs,
			"state":self.state,
			"clr":self.user_color,
			"user":self.user,
			"msg":self.message
		}
	
	@staticmethod
	def encode_message(msg):
		for k,v in ENCODE_DICT.items():
			msg = msg.replace(k, v)
		
		return msg
	
	@staticmethod
	def decode_message(msg):
		for k,v in DECODE_DICT.items():
			msg = msg.replace(k, v)
		
		return msg
	
	@staticmethod
	def encode_state(state):
		return CHAT_STATE.index(state)
	
	@staticmethod
	def decode_state(state):
		return CHAT_STATE[state]


def get_channels(channel_ids: List[str]) -> List[Channel]:
	"""
	Uses a (blocking) HTTP request to retrieve channel information from Twitch's API.

	:param channel_ids: A list of channel login name strings.
	:returns: A list of Channel objects.
	"""

	# Make channel objects and store them in a list
	channels = []
	for channel_id in channel_ids:
		query = gql.GET_CHANNEL_QUERY.format(channel_id=channel_id)
		resp = gql.gql_query(query=query).json()
		c = resp["data"]["user"]

		if c == None:
			raise Exception(f"Channel `{channel_id}` does not exist!")
		
		channels.append(
			Channel(
				id=c["id"], login=c["login"],
				display_name=c["displayName"], created_at=c["createdAt"]
			)
		)
	
	return channels


def get_channel_vods(channel: Channel) -> List[Vod]:
	"""
	Uses a (blocking) HTTP request to retrieve VOD info for a specific channel.

	:param channel: A Channel object.
	:returns: A list of VOD objects.
	"""

	vods = []
	pagination = ""
	while True:
		query = gql.GET_CHANNEL_VIDEOS_QUERY.format(
			channel_id=channel.login,
			after=pagination, first=100,
			type="ARCHIVE", sort="TIME"
		)
		resp = gql.gql_query(query=query).json()
		resp = resp ["data"]["user"]["videos"]

		if not resp or not resp["edges"]:
			break

		pagination = resp["edges"][-1]["cursor"]
		
		for vod in resp["edges"]:
			v = vod["node"]
			c = v["creator"]
			g = v["game"]

			game_id = ""
			game_name = ""
			if g:
				game_id = g["id"]
				game_name = g["name"]

			vods.append(
				Vod(
					id=v["id"], length=v["lengthSeconds"], title=v["title"],
					user_id=c["id"], user_login=c["login"], user_name=c["displayName"], 
					game_id=game_id, game_name=game_name, created_at=v["publishedAt"]
				)
			)

		if pagination == "" or pagination == None:
			break

	return vods


def get_channel_clips(channel: Channel) -> List[Clip]:
	"""
	Uses a (blocking) HTTP request to retrieve Clip info for a specific channel.

	:param channel: A Channel object.
	:returns: A list of Clip objects.
	"""

	clips = []
	pagination = ""
	while True:
		query = gql.GET_CHANNEL_CLIPS_QUERY.format(
			channel_id=channel.login,
			after=pagination, first=100
		)
		resp = gql.gql_query(query=query).json()
		resp = resp["data"]["user"]["clips"]

		if not resp or not resp["edges"]:
			break

		pagination = resp["edges"][-1]["cursor"]
		
		for clip in resp["edges"]:
			c = clip["node"]
			b = c["broadcaster"]
			w = c["curator"]
			g = c["game"]
			v = c["video"]
			v_id = "unknown"
			if v is not None:
				v_id = v["id"]

			w_id = b["id"]
			w_login = b["login"]
			w_name = b["displayName"]
			if w is not None:
				w_id = w["id"]
				w_login = w["login"]
				w_name = w["displayName"]
			
			g_id = ""
			g_name = ""
			if g is not None:
				g_id = g["id"]
				g_name = g["name"]

			clips.append(
				Clip(
					id=c["id"], slug=c["slug"], created_at=c["createdAt"],
					user_id=b["id"], user_login=b["login"], user_name=b["displayName"],
					clipper_id=w_id, clipper_login=w_login, clipper_name=w_name,
					game_id=g_id, game_name=g_name, title=c["title"],
					view_count=c["viewCount"], length=c["durationSeconds"],
					offset=c["videoOffsetSeconds"] or 0, video_id=v_id
				)
			)

		if pagination == "" or pagination == None:
			break

	return clips


def get_video_comments(video_id: str) -> List[ChatMessage]:
	"""
	Uses a (blocking) HTTP request to retrieve chat logs for a specific video.

	:param video_id: A video ID string.
	:returns: A list of ChatMessage objects.
	"""

	messages = []
	pagination = ""
	while True:
		query = gql.GET_VIDEO_COMMENTS_QUERY.format(
			video_id=video_id, first=100, after=pagination
		)
		resp = gql.gql_query(query=query).json()
		resp = resp["data"]["video"]["comments"]

		if not resp or not resp["edges"]:
			break

		pagination = resp["edges"][-1]["cursor"]
		for comment in resp["edges"]:
			c = comment["node"]

			usr = c["commenter"]["displayName"]
			clr = c["message"]["userColor"] or "FFFFFF"
			clr = clr.strip("#")

			msg = ""
			for frag in c["message"]["fragments"]:
				if frag["mention"] is not None:
					msg += "@" + frag["mention"]["displayName"]
				msg += frag["text"]
				msg += " "
			msg = msg.strip()

			ofst = c["contentOffsetSeconds"]
			state = c["state"]

			messages.append(
				ChatMessage(
					user=usr, color=clr, msg=msg, offset=ofst, state=state
				)
			)
		
		if pagination == "" or pagination == None:
			break
	
	return messages



# Module to pull and create different files and directories on the OS

from .printer import cprint

import os
import sys
import json
from pathlib import Path
from collections import OrderedDict


vodbotdir = Path.home() / ".vodbot"
DEFAULT_CONF_PATH = vodbotdir / "conf.json"
DEFAULT_CONF = OrderedDict([
	("twitch_channels", []), # channels to watch for new clips and videos

	("pull_chat_logs", True), # determines if chat logs get pulled with VODs

	("stage_timezone", "+0000"), # timezone for when a video happened
	("stage_format", { # Macros for video descriptions when staging
		"watch": "-- Watch live at {links}",
		"discord": "-- Join the Discord at https://discord.gg/v2t6uag",
		"credits": "\n{watch}\n{discord}"
	}),
	("stage_upload_delete", True), # delete a stage on completed upload?
	("stage_export_delete", True), # delete a stage on completed export?

	("chat_msg_time", 10), # measured in seconds, how long to show a chat message in subtitle
	("chat_upload", "RealText"), # can be RealText or SAMI
	("chat_export", "raw"), # can be raw, RealText, or SAMI

	("ffmpeg_loglevel", "warning"), # warning (recommended), error (only breaking stuff), fatal (absolute error)
	
	("youtube_client_path", str(vodbotdir / "yt-client.json")), # google generated json for using youtube api
	("youtube_pickle_path", str(vodbotdir / "yt-session.pkl")), # caching oauth session and info
	
	("vod_dir", str(vodbotdir / "vods")), # place to store video data
	("clip_dir", str(vodbotdir / "clips")), # clip data
	("temp_dir", str(vodbotdir / "temp")), # temp data
	("stage_dir", str(vodbotdir / "stage")) # stage data
])


def make_dir(directory):
	"""
	Creates the directory structure to house the configuration data and VODs.

	:param directory: A string of where the directory structure should be made.
	"""
	os.makedirs(str(directory), exist_ok=True)


def load_conf(filename):
	"""
	Loads the config of VodBot at a specific directory.

	:param filename: File name of the JSON formatted configuration file.
	:returns: Dictionary created from the conf file.
	"""

	conf = None
	try:
		with open(filename) as f:
			conf = json.load(f)
	except FileNotFoundError:
		exit_prog(2, f"Config not found. You can configure VodBot with the init command.")
	except json.decoder.JSONDecodeError as e:
		exit_prog(98, f"Failed to decode config. \"{e.msg}\"")
		
	for key in DEFAULT_CONF:
		if key not in conf:
			exit_prog(79, f"Missing key \"{key}\" in config, please edit your config to continue.")
	
	if conf["youtube_client_path"] == "":
		cprint("Please edit your config with your Youtube Client ID and Secret to use the upload command.")

	if len(conf["twitch_channels"]) == 0:
		exit_prog(40, "No channels listed in config, please edit your config to continue.")

	chat_format = ["RealText", "SAMI"]
	if conf["chat_upload"] not in chat_format:
		exit_prog(10, f"Chat format for uploading not valid. Got `{conf['chat_upload']}`, expected any of the following `{chat_format}`. Fix your config to continue.")
	chat_format.append("raw")
	if conf["chat_export"] not in chat_format:
		exit_prog(10, f"Chat format for exporting not valid. Got `{conf['chat_export']}`, expected any of the following `{chat_format}`. Fix your config to continue.")
	
	# TODO: Check all important config options

	return conf


def exit_prog(code=0, errmsg=None):
	"""
	Exits the program with an error code and optional error message.

	:param code: The error code to exit with. Should be unique per exit case.
	:param errmsg: The corresponding error message to print when exiting.
	"""

	print()

	if code != 0:
		msg = f"#r#fR#lERROR! #fY#l({code})#r"
		if errmsg != None:
			msg += " #fR#l" + errmsg + "#r"
		cprint(msg, end=" ")

	cprint("#r#dExiting...#r")
	sys.exit(code)



# Module to call GQL queries

from vodbot import TWITCH_CLIENT_ID

import requests
from urllib.parse import urlencode


GQL_URL = "https://gql.twitch.tv/gql"
GQL_HEADERS = {"Client-ID": TWITCH_CLIENT_ID} # We use Twitch's private client ID for GQL calls


def _process_query_errors(resp):
	if 400 <= resp.status_code < 500:
		data = resp.json()
		raise Exception(data["message"]) # TODO: make exceptions?
	resp.raise_for_status()

	if "errors" in resp:
		raise Exception(resp["errors"])


def gql_get(params={}):
	resp = requests.get(GQL_URL, params=params, headers=GQL_HEADERS)
	_process_query_errors(resp)
	return resp


def gql_post(data=None, json=None):
	resp = requests.post(GQL_URL, data=data, json=json, headers=GQL_HEADERS)
	_process_query_errors(resp)
	return resp


def gql_query(query=None, data=None):
	resp = requests.post(GQL_URL, json={"query":query}, data=data, headers=GQL_HEADERS)
	_process_query_errors(resp)
	return resp


# GQL Query forms
# Channel VODs query
GET_CHANNEL_VIDEOS_QUERY = """
{{  user(login: "{channel_id}") {{
		videos( first: {first}, type: {type}, sort: {sort}, after: "{after}" ) {{
			totalCount
			edges {{ cursor
				node {{ id title
					publishedAt broadcastType
					lengthSeconds thumbnailURLs
					game {{ id name }}
					creator {{ id login displayName }}
}}  }}  }}  }}  }}
"""
# Channel Clips query
# period can be LAST_DAY, LAST_WEEK, LAST_MONTH, or ALL_TIME
GET_CHANNEL_CLIPS_QUERY = """
{{  user(login: "{channel_id}") {{
		clips(
			first: {first}, after: "{after}",
			criteria: {{ period: ALL_TIME, sort: VIEWS_DESC }}
		) {{
			edges {{ cursor
				node {{ id slug title
					createdAt viewCount
					durationSeconds url videoOffsetSeconds
					video {{ id }}
					game {{ id name }}
					broadcaster {{ id displayName login }}
					curator {{ id displayName login }}
}}  }}  }}  }}  }}
"""
# Single VOD query
GET_VIDEO_QUERY = """
{{ video(id: "{video_id}") {{
		id title publishedAt
		broadcastType lengthSeconds
		game {{ id name }}
		creator {{ id login displayName }}
}}  }}
"""
# Single Clip query
GET_CLIP_QUERY = """
{{  clip(slug: "{clip_slug}") {{
		videoQualities {{ frameRate quality sourceURL }}
}}  }}
"""
# Single Channel info query
GET_CHANNEL_QUERY = """
{{  user(login: "{channel_id}") {{
		id login displayName
		description createdAt
		roles {{ isAffiliate isPartner }}
		stream {{
			id title type
			viewersCount createdAt
			game {{ id name }}
}}  }}  }}
"""
# IRC Chat query
GET_VIDEO_COMMENTS_QUERY = """
{{ video(id: "{video_id}") {{
	comments(contentOffsetSeconds: 0, after: "{after}") {{
		edges {{ cursor node {{
			contentOffsetSeconds state
			commenter {{ displayName }}
			message {{ userColor 
				fragments {{ mention {{ displayName }} text }}
}}  }}  }}  }}  }} }}
"""


VIDEO_ACCESS_QUERY = """
{{  videoPlaybackAccessToken(
		id: {video_id},
		params: {{
			platform: "web",
			playerBackend: "mediaplayer",
			playerType: "site"
		}}
	) {{ signature value }}
}}
"""

def get_access_token(video_id):
	"""
	Gets the VOD player access token, used to stream the VOD to the host.
	"""
	query = VIDEO_ACCESS_QUERY.format(video_id=video_id)

	resp = gql_query(query=query).json()

	return resp["data"]["videoPlaybackAccessToken"]


def get_clip_source(clip_slug):
	query = GET_CLIP_QUERY.format(clip_slug=clip_slug)

	resp = gql_query(query=query).json()

	url = resp["data"]["clip"]["videoQualities"][0]["sourceURL"]

	query = {
		"operationName": "VideoAccessToken_Clip",
		"variables": {
			"slug": clip_slug
		},
		"extensions": {
			"persistedQuery": {
				"version": 1,
				"sha256Hash": "36b89d2507fce29e5ca551df756d27c1cfe079e2609642b4390aa4c35796eb11"
			}
		}
	}

	resp = gql_post(json=query).json()
	token = resp["data"]["clip"]["playbackAccessToken"]
	token = urlencode({
		"sig": token["signature"],
		"token": token["value"]
	})

	url = f"{url}?{token}"

	return url



from sys import argv as sys_argv

# Taken from https://github.com/tartley/colorama#recognised-ansi-sequences
COLOR_CODES = {
	"r": "\033[0m", # reset
	"c": "\033[K",  # clear line
	"l": "\033[1m", # light
	"d": "\033[2m", # dim
	"A": "\033[A",  # move up a line
	"F": "\033[F",  # move cursor to beginning of line

	"fK": "\033[30m", # foreground blacK
	"fR": "\033[31m", # Red
	"fG": "\033[32m", # Green
	"fY": "\033[33m", # Yellow
	"fB": "\033[34m", # Blue
	"fM": "\033[35m", # Magenta
	"fC": "\033[36m", # Cyan
	"fW": "\033[37m", # White

	"bK": "\033[40m", # background blacK
	"bR": "\033[41m",
	"bG": "\033[42m",
	"bY": "\033[43m",
	"bB": "\033[44m",
	"bM": "\033[45m",
	"bC": "\033[46m",
	"bW": "\033[47m"
}

USE_COLOR = "--no-color" not in sys_argv

def colorize(text: str):
	for k, v in COLOR_CODES.items():
		text = text.replace("#" + k, v)
	return text

def strip_color(text: str):
	for k in COLOR_CODES:
		text = text.replace("#" + k, "")
	return text

def cprint(*args, **kwargs):
	args = [colorize(txt) if USE_COLOR else strip_color(txt) for txt in args]
	print(*args, **kwargs)


# Upload, command to upload staged videos to YouTube
# References:
# https://developers.google.com/youtube/v3/guides/uploading_a_video
# https://learndataanalysis.org/how-to-upload-a-video-to-youtube-using-youtube-data-api-in-python/

from .stage import StageData

import vodbot.util as util
import vodbot.video as vbvid
import vodbot.chatlog as vbchat
from vodbot.printer import cprint

import json
import pickle
from datetime import datetime
from pathlib import Path
from os import remove as os_remove
from os.path import exists as os_exists
from time import sleep

from httplib2.error import HttpLib2Error, HttpLib2ErrorWithResponse

from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.discovery import build
from googleapiclient.http import MediaFileUpload
from googleapiclient.errors import HttpError, ResumableUploadError
from google.auth.transport.requests import Request
from google.auth.exceptions import RefreshError


# Default path
vodbotdir = util.vodbotdir
stagedir = None
tempdir = None

RETRIABLE_EXCEPTS = (HttpLib2Error, HttpLib2ErrorWithResponse, IOError)


EPOCH = datetime.utcfromtimestamp(0)
def sort_stagedata(stagedata):
	date = datetime.strptime(stagedata.datestring, "%Y/%m/%d")
	return (date - EPOCH).total_seconds()


def upload_video(conf: dict, service, stagedata: StageData) -> str:
	tmpfile = None
	try:
		tmpfile = vbvid.process_stage(conf, stagedata)
	except vbvid.FailedToSlice as e:
		cprint(f"#r#fRSkipping stage `{stagedata.id}`, failed to slice video with ID of `{e.vid}`.#r\n")
	except vbvid.FailedToConcat:
		cprint(f"#r#fRSkipping stage `{stagedata.id}`, failed to concatenate videos.#r\n")
	except vbvid.FailedToCleanUp as e:
		cprint(f"#r#fRSkipping stage `{stagedata.id}`, failed to clean up temp files.#r\n\n{e.vid}")

	# send request to youtube to upload
	request_body = {
		"snippet": {
			"categoryId": 20,
			"title": stagedata.title,
			"description": stagedata.desc
		},
		"status": {
			"privacyStatus": "private",
			"selfDeclaredMadeForKids": False
		}
	}

	# create media file, 100 MiB chunks
	media_file = MediaFileUpload(str(tmpfile), chunksize=1024*1024*100, resumable=True)

	# create upload request and execute
	response_upload = service.videos().insert(
		part="snippet,status",
		body=request_body,
		notifySubscribers=False,
		media_body=media_file
	)

	video_id = None # youtube video id
	resp = None
	errn = 0
	cprint(f"#fCUploading stage #r`#fM{stagedata.id}#r`, progress: #fC0#fY%#r #d...#r", end="\r")
	
	# Below is taken from Google's documentation, rewrite?
	while resp is None:
		try:
			status, resp = response_upload.next_chunk()
			if status:
				cprint(f"#fCUploading stage #r`#fM{stagedata.id}#r`, progress: #fC{int(status.progress()*100)}#fY%#r #d...#r", end="\r")
			if resp is not None:
				if "id" in resp:
					cprint(f"#fCUploading stage #r`#fM{stagedata.id}#r`, progress: #fC100#fY%#r!")
					cprint(f"#l#fGVideo was successfully uploaded!#r #dhttps://youtu.be/{resp['id']}#r")
					video_id = resp["id"]
				else:
					util.exit_prog(99, f"Unexpected upload failure occurred, \"{resp}\"")
		except ResumableUploadError as err:
			if err.resp.status in [400, 401, 402, 403]:
				try:
					jsondata = json.loads(err.content)['error']['errors'][0]
					util.exit_prog(40, f"API Error: `{jsondata['reason']}`. Message: `{jsondata['message']}`")
				except (json.JSONDecodeError, KeyError):
					util.exit_prog(40, f"Unknown API Error has occured, ({err.resp.status}, {err.content})")
			print(f"A Resumeable error has occured, retrying in 5 sec... ({err.resp.status}, {err.content})")
			errn += 1
			sleep(5)
		except HttpError as err:
			if err.resp.status in [500, 502, 503, 504]:
				print(f"An HTTP error has occured, retrying in 5 sec... ({err.resp.status}, {err.content})")
				errn += 1
				sleep(5)
			else:
				raise
		except RETRIABLE_EXCEPTS as err:
			print(f"An HTTP error has occured, retrying in 5 sec... ({err})")
			errn += 1
			sleep(5)
		
		if errn >= 10:
			print("Skipping video upload, errored too many times.")
			break
	
	# we're done, lets clean up
	else:
		try:
			# delete vars to release the files
			del media_file
			del response_upload
			sleep(1)
			os_remove(str(tmpfile))
		except Exception as e:
			util.exit_prog(90, f"Failed to remove temp slice file of stage `{stagedata.id}` after upload. {e}")
	
	return video_id


def upload_captions(conf: dict, service, stagedata: StageData, vid_id: str) -> bool:
	tmpfile = vbchat.process_stage(conf, stagedata, "upload")

	if tmpfile == False:
		return True

	request_body = {
		"snippet": {
			"name": "Chat",
			"videoId": vid_id,
			"language": "en"
		}
	}

	media_file = MediaFileUpload(str(tmpfile), chunksize=1024*1024*100, resumable=True)

	response_upload = service.captions().insert(
		part="snippet",
		body=request_body,
		sync=False,
		media_body=media_file
	)

	resp = None
	errn = 0
	cprint(f"#fCUploading stage chatlog #r`#fM{stagedata.id}#r`, progress: #fC0#fY%#r #d...#r", end="\r")
	
	while resp is None:
		try:
			status, resp = response_upload.next_chunk()
			if status:
				cprint(f"#fCUploading stage chatlog #r`#fM{stagedata.id}#r`, progress: #fC{int(status.progress()*100)}#fY%#r #d...#r", end="\r")
			if resp is not None:
				cprint(f"#fCUploading stage chatlog #r`#fM{stagedata.id}#r`, progress: #fC100#fY%#r!")
				cprint(f"#l#fGVideo captions were successfully uploaded!#r")
		except ResumableUploadError as err:
			if err.resp.status in [400, 401, 402, 403]:
				try:
					jsondata = json.loads(err.content)['error']['errors'][0]
					util.exit_prog(40, f"API Error: `{jsondata['reason']}`. Message: `{jsondata['message']}`")
				except (json.JSONDecodeError, KeyError):
					util.exit_prog(40, f"Unknown API Error has occured, ({err.resp.status}, {err.content})")
			print(f"A Resumeable error has occured, retrying in 5 sec... ({err.resp.status}, {err.content})")
			errn += 1
			sleep(5)
		except HttpError as err:
			if err.resp.status in [500, 502, 503, 504]:
				print(f"An HTTP error has occured, retrying in 5 sec... ({err.resp.status}, {err.content})")
				errn += 1
				sleep(5)
			else:
				raise
		except RETRIABLE_EXCEPTS as err:
			print(f"An HTTP error has occured, retrying in 5 sec... ({err})")
			errn += 1
			sleep(5)
		
		if errn >= 10:
			print("Skipping chatlog upload, errored too many times.")
			return False
	
	# we're done, lets clean up
	else:
		try:
			# delete vars to release the files
			del media_file
			del response_upload
			sleep(1)
			os_remove(str(tmpfile))
		except Exception as e:
			util.exit_prog(90, f"Failed to remove temp slice file of stage `{stagedata.id}` after upload. {e}")
	
	return True


def run(args):
	# load config
	conf = util.load_conf(args.config)

	# configure variables
	global stagedir, tempdir
	tempdir = Path(conf["temp_dir"])
	stagedir = Path(conf["stage_dir"])
	PICKLE_FILE = conf["youtube_pickle_path"]
	CLIENT_SECRET_FILE = conf["youtube_client_path"]
	API_NAME = 'youtube'
	API_VERSION = 'v3'
	SCOPES = [ # Only force-ssl is required, but both makes it explicit.
		"https://www.googleapis.com/auth/youtube.upload",   # Videos.insert
		"https://www.googleapis.com/auth/youtube.force-ssl" # Captions.insert
	]

	# handle logout
	if args.id == "logout":
		try:
			os_remove(PICKLE_FILE)
			cprint("#dLogged out of Google API session#r")
		except:
			util.exit_prog(11, "Failed to remove credentials for YouTube account.")
		
		return
	
	# load stages, but dont upload
	# Handle id/all
	stagedata = None
	stagedatas = None
	if args.id == "all":
		cprint("#dLoading stages...", end=" ")
		# create a list of all the hashes and sort by date streamed, upload chronologically
		stagedatas = StageData.load_all_stages(stagedir)
		stagedatas.sort(key=sort_stagedata)
	else:
		cprint("#dLoading stage...", end=" ")
		# check if stage exists, and prep it for upload
		stagedata = StageData.load_from_id(stagedir, args.id)
		cprint(f"About to upload stage {stagedata.id}.#r")

	# authenticate youtube service
	if not os_exists(CLIENT_SECRET_FILE):
		util.exit_prog(19, "Missing YouTube Client ID/Secret file.")

	cprint("Authenticating with Google...", end=" ")

	service = None
	credentials = None

	if os_exists(PICKLE_FILE):
		with open(PICKLE_FILE, "rb") as f:
			credentials = pickle.load(f)
	
	if not credentials or credentials.expired:
		try:
			if credentials and credentials.expired and credentials.refresh_token:
				credentials.refresh(Request())
			else:
				flow = InstalledAppFlow.from_client_secrets_file(CLIENT_SECRET_FILE, SCOPES)
				credentials = flow.run_console()
		except RefreshError:
			flow = InstalledAppFlow.from_client_secrets_file(CLIENT_SECRET_FILE, SCOPES)
			credentials = flow.run_console()
		
		with open(PICKLE_FILE, "wb") as f:
			pickle.dump(credentials, f)
	
	try:
		service = build(API_NAME, API_VERSION, credentials=credentials)
	except Exception as err:
		util.exit_prog(50, f"Failed to connect to YouTube API, \"{err}\"")
	
	cprint("Authenticated.", end=" ")
	
	# Handle id/all
	if args.id == "all":
		# begin to upload
		cprint(f"About to upload {len(stagedatas)} stages.#r")
		for stage in stagedatas:
			video_id = upload_video(conf, service, stage)
			if video_id is not None:
				chat_success = True
				if conf["chat_upload"]:
					chat_success = upload_captions(conf, service, stagedata, video_id)
				
				if conf["stage_upload_delete"] and chat_success:
					try:
						os_remove(str(stagedir / f"{stage.id}.stage"))
					except:
						util.exit_prog(90, f"Failed to remove stage `{stage.id}` after upload.")
			print()
	else:
		# upload stage
		cprint(f"About to upload stage {stagedata.id}.#r")
		video_id = upload_video(conf, service, stagedata)
		if video_id is not None:
			chat_success = True
			if conf["chat_upload"]:
				chat_success = upload_captions(conf, service, stagedata, video_id)
			
			if conf["stage_upload_delete"] and chat_success:
				try:
					os_remove(str(stagedir / f"{stagedata.id}.stage"))
				except:
					util.exit_prog(90, f"Failed to remove stage `{stagedata.id}` after upload.")
	



# Info, command to inspect a piece of media from Twitch

import re
from vodbot.printer import cprint

from vodbot import util
from vodbot.itd import gql

PATTERNS = {
	"vod": [
		r"^(?P<id>\d+)?$",
		r"^(https?://)?(www\.)?twitch.tv/videos/(?P<id>\d+)(\?.*)?$"
	],
	"clip": [
		r"^(?P<id>[A-Za-z0-9]+(?:-[A-Za-z0-9_-]{16})?)$",
		r"^(https?://)?(www\.)?twitch.tv/\w+/clip/(?P<id>[A-Za-z0-9]+(?:-[A-Za-z0-9_-]{16})?)(\?.*)?$",
		r"^(https?://)?clips\.twitch.tv/(?P<id>[A-Za-z0-9]+(?:-[A-Za-z0-9_-]{16})?)(\?.*)?$"
	],
	"channel": [
		r"^[a-zA-Z0-9][\w]{0,24}$",
		r"^(https?://)?(www\.)?twitch\.tv/(?P<id>[a-zA-Z0-9][\w]{0,24})(\?.*)?$"
	]
}

KEYTRANS = {
	"vod": "VOD",
	"clip": "Clip",
	"channel": "Channel"
}


def get_type(cid):
	for ctype, clist in PATTERNS.items():
		for pattern in clist:
			match = re.match(pattern, cid)
			if match:
				return match.group("id"), ctype
	
	return None, None


def run(args):
	# First determine what kind of data it is...
	# We should allow people to paste full links or just the ID too.
	tid = args.id

	# Get the proper id and type of content we need to pull
	cprint("#dDetermining type...#r", end=" ")
	cid, ctype = get_type(tid)

	# The call the appropriate info query for GQL
	query = None
	if ctype == "vod":
		query = gql.GET_VIDEO_QUERY.format(video_id=cid)
	elif ctype == "clip":
		query = gql.GET_CLIP_QUERY.format(clip_slug=cid)
	elif ctype == "channel":
		query = gql.GET_CHANNEL_QUERY.format(channel_id=cid)
	else:
		util.exit_prog(92, "Could not determine content type from input.")
	cword = KEYTRANS[ctype]
		
	cprint(f"#dQuerying {cword} content for `{cid}`...#r")
	
	# run the query
	resp = gql.gql_query(query=query).json()
	resp = resp["data"]
	if not any([resp.get("video"), resp.get("clip"), resp.get("user")]):
		util.exit_prog(93, f"Could not query info on {cword} content from input.")
	
	# Read the data back and out to the terminal
	if ctype == "vod":
		r = resp['video']
		c = r['creator']
		g = r['game']
		cprint(f"#fGTitle: `{r['title']}`#r - ID: `{r['id']}`")
		cprint(f"#fMBroadcaster: {c['displayName']} - {c['login']} ({c['id']})#r")
		cprint(f"#fCPlaying: {g['name']} ({g['id']})#r")
		cprint(f"#fYAt: {r['publishedAt']} - For: {r['lengthSeconds']} seconds#r")
	elif ctype == "clip":
		r = resp['clip']
		c = r['curator']
		b = r['broadcaster']
		g = r['game']
		cprint(f"#fGTitle: `{r['title']}`#r - ID: `{r['id']}`")
		cprint(f"Slug: `{r['slug']}`")
		cprint(f"#fBClipper: {c['displayName']} - {c['login']} ({c['id']})#r")
		cprint(f"#fMBroadcaster: {b['displayName']} - {b['login']} ({b['id']})#r")
		cprint(f"#fCPlaying: {g['name']} ({g['id']})#r")
		cprint(f"#fYAt: {r['createdAt']} - For: {r['durationSeconds']} seconds - With: {r['viewCount']} views#r")
	elif ctype == "channel":
		r = resp['user']
		cprint(f"#fMBroadcaster: {r['displayName']} - {r['login']} (ID: `{r['id']}`)#r")
		cprint(f"#fRDescription: {r['description']}#r")
		cprint(f"#fYChannel Created At: {r['createdAt']}#r")
		r = r['roles']
		cprint(f"#fBRoles: Affiliate={r['isAffiliate']} - Partner={r['isPartner']}#r")
		#cprint(f"Site Roles: Staff={r['isStaff']} - GlobalMod={r['isGlobalMod']} - SiteAdmin={r['isSiteAdmin']}")

	# Done!



from pathlib import Path

from . import gql, worker
from vodbot import chatlog
from vodbot.util import make_dir, vodbotdir
from vodbot.printer import cprint
from vodbot.twitch import Vod, Clip, ChatMessage, get_video_comments

import subprocess
import requests
import shutil
import m3u8
import os
import json

class JoiningFailed(Exception):
	pass

def get_playlist_uris(video_id, access_token):
	"""
	Grabs the URI's for accessing each of the video chunks.
	"""
	url = f"http://usher.twitch.tv/vod/{video_id}"

	resp = requests.get(url, params={
		"nauth": access_token['value'],
		"nauthsig": access_token['signature'],
		"allow_source": "true",
		"player": "twitchweb",
	})
	resp.raise_for_status()

	data = resp.content.decode("utf-8")

	playlist = m3u8.loads(data)
	playlist_uris = []

	for p in playlist.playlists:
		playlist_uris += [p.uri]
	
	return playlist_uris

def dl_video(video: Vod, TEMP_DIR: Path, path: str, max_workers: int, LOG_LEVEL: str):
	video_id = video.id

	# Grab access token
	access_token = gql.get_access_token(video_id)

	# Get M3U8 playlist, and parse them
	# (first URI is always source quality!)
	uris = get_playlist_uris(video_id, access_token)
	source_uri = uris[0]

	# Fetch playlist at proper quality
	resp = requests.get(source_uri)
	resp.raise_for_status()
	playlist = m3u8.loads(resp.text)

	# Create a temp dir in .vodbot/temp
	tempdir = TEMP_DIR / video_id
	make_dir(str(tempdir))

	# Dump playlist to a file
	playlist_path = tempdir / "playlist.m3u8"
	playlist.dump(str(playlist_path))

	# Get all the necessary vod paths for the uri
	base_uri = "/".join(source_uri.split("/")[:-1]) + "/"
	vod_paths = []
	for segment in playlist.segments:
		if segment.uri not in vod_paths:
			vod_paths.append(segment.uri)

	# Download VOD chunks to the temp folder
	path_map = worker.download_files(video_id, base_uri, tempdir, vod_paths, max_workers)
	# TODO: rewrite this output to look nicer and remove ffmpeg output using COLOR_CODES["F"]
	cprint("#dDone, now to ffmpeg join...#r")

	# join the vods using ffmpeg at specified path
	# TODO: change this to the concat function in video?
	cwd = os.getcwd()
	os.chdir(str(tempdir))
	cmd = [
		"ffmpeg", "-i", str(playlist_path),
		"-c", "copy", path, "-y",
		"-stats", "-loglevel", LOG_LEVEL
	]
	result = subprocess.run(cmd)
	os.chdir(cwd)

	if result.returncode != 0:
		raise JoiningFailed()

	# delete temp folder and contents
	shutil.rmtree(str(tempdir))


def dl_video_chat(video: Vod, path: str):
	video_id = video.id

	# Download all chat from video
	cprint(f"#fM#lVOD Chat#r `#fM{video_id}#r` (0%)", end="")
	msgs = get_video_comments(video_id)
	cprint(f"\r#fM#lVOD Chat#r `#fM{video_id}#r` (100%); Done, now to write...", end="")

	chatlog.chat_to_logfile(msgs, path)

	cprint(f"\r#fM#lVOD Chat#r `#fM{video_id}#r` (100%); Done, now to write... Done")


def dl_clip(clip: Clip, path: str):
	clip_slug = clip.slug
	clip_id = clip.id

	# Get proper clip file URL
	source_url = gql.get_clip_source(clip_slug)

	# Download file to path
	size, _existed = worker.download_file(source_url, path)

	# Print progress
	cprint(f"#fM#lClip#r ({clip_id})`#fM{clip_slug}#r` #fB#l~{worker.format_size(size)}#r")


# Export video data to a specific location

from .stage import StageData

import vodbot.util as util
import vodbot.video as vbvid
import vodbot.chatlog as vbchat
from vodbot.printer import cprint

from datetime import datetime
from pathlib import Path
from os import remove as os_remove, replace as os_replace


# Default path
vodbotdir = util.vodbotdir
stagedir = None


DISALLOWED_CHARACTERS = [
	"/", "\\", "<", ">", ":",
	"\"", "|", "?", "*"
]


EPOCH = datetime.utcfromtimestamp(0)
def sort_stagedata(stagedata):
	date = datetime.strptime(stagedata.datestring, "%Y/%m/%d")
	return (date - EPOCH).total_seconds()


def handle_stage(conf: dict, stage: StageData) -> Path:
	tmpfile = None
	try:
		tmpfile = vbvid.process_stage(conf, stage)
	except vbvid.FailedToSlice as e:
		cprint(f"#r#fRSkipping stage `{stage.id}`, failed to slice video with ID of `{e.vid}`.#r\n")
		return None
	except vbvid.FailedToConcat:
		cprint(f"#r#fRSkipping stage `{stage.id}`, failed to concatenate videos.#r\n")
		return None
	except vbvid.FailedToCleanUp as e:
		cprint(f"#r#fRSkipping stage `{stage.id}`, failed to clean up temp files.#r\n\n{e.vid}")
		return None
	
	return tmpfile


def run(args):
	global stagedir

	conf = util.load_conf(args.config)
	stagedir = Path(conf["stage_dir"])

	util.make_dir(args.path)
	args.path = Path(args.path)
	
	# load stages, but dont slice
	# Handle id/all
	if args.id == "all":
		cprint("#dLoading and slicing stages...#r")

		# create a list of all the hashes and sort by date streamed, slice chronologically
		stagedatas = StageData.load_all_stages(stagedir)
		stagedatas.sort(key=sort_stagedata)

		for stage in stagedatas:
			# Export with ffmpeg
			tmpfile = handle_stage(conf, stage)
			# Export chat
			tmpchat = vbchat.process_stage(conf, stage, "export")

			title = stage.title.strip()
			for x in DISALLOWED_CHARACTERS:
				title = title.replace(x, "_")

			# move appropriate files
			if tmpfile is not None:
				os_replace(tmpfile, args.path / (title+tmpfile.suffix))
			if tmpchat is not None:
				os_replace(tmpchat, args.path / (title+tmpchat.suffix))
			
			# deal with old stage
			if conf["stage_export_delete"]:
				os_remove(str(stagedir / f"{stage.id}.stage"))
	else:
		cprint("#dLoading stage...", end=" ")
		
		# check if stage exists, and prep it for slice
		stagedata = StageData.load_from_id(stagedir, args.id)
		
		# Export with ffmpeg
		tmpfile = handle_stage(conf, stagedata)
		# Export chat
		tmpchat = vbchat.process_stage(conf, stagedata, "export")
		
		title = stagedata.title.strip()
		for x in DISALLOWED_CHARACTERS:
			title = title.replace(x, "_")

		# move appropriate files
		if tmpfile is not None:
			os_replace(tmpfile, args.path / (title+tmpfile.suffix))
		if tmpchat is not None:
			os_replace(tmpchat, args.path / (title+tmpchat.suffix))
		
		# Deal with old stage
		if conf["stage_export_delete"]:
			os_remove(str(stagedir / f"{stagedata.id}.stage"))

	# say "Done!"
	cprint("#fG#lDone!#r")



# Module that parses chatlogs to and from files

from os import write
from vodbot.commands.stage import StageData
from vodbot.printer import cprint
from vodbot.twitch import ChatMessage
import vodbot.util as util

import json
from typing import List
from pathlib import Path
import shutil


def chat_to_logfile(msgs: List[ChatMessage], path: str) -> None:
	# create preamble (contains all chatter's names and their colors)
	preamb = []
	for m in msgs:
		s = f"{m.color};{m.user}"
		if s not in preamb:
			preamb.append(s)

	# open chat log file
	with open(path, "wb") as f:
		# write pramble as one line, delimited by null characters
		for s in preamb:
			f.write(s.encode("utf-8"))

			if s != preamb[-1]:
				f.write("\0".encode("utf-8"))
		
		# preamble done
		f.write("\n".encode("utf-8"))

		# write the messages, each line a new message
		for m in msgs:
			idx = preamb.index(f"{m.color};{m.user}")
			# each line is a unique message
			# line string split with \0 for components
			# m[0]=offset from start of stream, m[1]=state (published, deleted, etc),
			# m[2]=user (in preamble), m[3:]=message they sent, encoded as a string
			f.write(f"{m.offset}\0{m.enc_state}\0{idx}\0{m.enc_msg}".encode("utf-8"))

			# newlines on every line except the last
			if m != msgs[-1]:
				f.write("\n".encode("utf-8"))


def logfile_to_chat(path: str) -> List[ChatMessage]:
	chats = []

	with open(path, "rb") as f:
		readfirst = False
		preamb = []
		for line in f.readlines():
			line = line.decode("utf-8").strip("\n") # Remove newline character and decode as utf8
			if not readfirst:
				# first line is preamble, so we unravel it
				for user in line.split("\0"):
					user = user.split(";")
					preamb.append((user[0], user[1]))
				readfirst = True
			else:
				# succeeding lines are all chat messages, so we read each one, match it with a user
				# from the preamble, grab all the other info, make a ChatMessage object, and tack it
				# on to the list.
				line = line.split("\0")
				info = int(line[2])
				chats.append(
					ChatMessage(
						user=preamb[info][1], color=preamb[info][0],
						offset=int(line[0]), enc_state=int(line[1]), enc_msg=line[3]
					)
				)

	return chats


def chat_to_listwithbounds(msgs: List[ChatMessage], vid_duration:int, msg_duration:int) -> List[dict]:
	# First we run through all the messages and see what needs showing and what doesn't
	chat_lists = []
	# this is so cursed
	last_msgs: List[ChatMessage] = []
	for t in range(vid_duration):
		current_msgs: List[ChatMessage] = []
		for m in msgs:
			if t <= m.offset <= t+msg_duration:
				current_msgs.append(m)
			elif m.offset > t+msg_duration:
				break
		
		if current_msgs != last_msgs:
			# write line and overwrite last_msgs
			last_msgs = current_msgs
			write = {"begin": t, "end": vid_duration, "msgs":[], "break":False}
			if len(current_msgs) != 0:
				# write full messages line
				for m in current_msgs:
					# write color, name, and message
					write["msgs"] += [{"clr":m.color, "usr":m.user, "msg":m.msg}]
			else:
				# write clear line
				write["break"] = True

			# write string
			chat_lists.append(write)
	
	# set end boundaries
	for x in range(len(chat_lists)):
		if x != len(chat_lists)-1:
			chat_lists[x]["end"] = chat_lists[x+1]["begin"]
	
	return chat_lists


def chat_to_realtext(msgs: List[ChatMessage], path: str, vid_duration:int, msg_duration:int):
	# get chat with message in bounds
	chat_lists = chat_to_listwithbounds(msgs, vid_duration, msg_duration)

	# write actual realtext stuff
	with open(path, "w") as f:
		# add preamble (stuff like opening tags)
		f.write('<window type="generic" wordwrap="true"><font color="#ffffff" text-align="left">\n')

		for c in chat_lists:
			f.write(f'<time begin="{c["begin"]}" end="{c["end"]}">')
			if c["break"]:
				f.write("<clear/>\n")
				continue
			for m in c["msgs"]:
				msg = m["msg"].replace("<", "&lt;").replace(">", "&gt;").replace("&", "&amp;")
				f.write(f'<font color="#{m["clr"]}"><b>{m["usr"]}</b></font>: {msg}')
				if m != c["msgs"][-1]:
					f.write("<br/>")
			
			f.write("\n")
		
		# finish up
		f.write("</font></window>")


def chat_to_sami(msgs: List[ChatMessage], path: str, vid_duration:int, msg_duration:int):
	# get chat with message in bounds
	chat_lists = chat_to_listwithbounds(msgs, vid_duration, msg_duration)

	with open(path, "w") as f:
		# write preamble stuff
		f.write("<SAMI><head><SAMIParam>\n")
		f.write(f"Metrics {{time:ms; duration: {vid_duration*1000};}}\nSpec {{MSFT:1.0;}}\n")
		f.write('</SAMIParam><style type="text/css"><!--\n')
		f.write(f"p {{text-align: left; color: white}}\n")
		f.write("--></style></head><body>\n")

		for c in chat_lists:
			f.write(f"<sync start={c['begin']*1000}><p>")
			for m in c["msgs"]:
				msg = m["msg"].replace("<", "&lt;").replace(">", "&gt;").replace("&", "&amp;")
				f.write(f'<font color="#{m["clr"]}"><b>{m["usr"]}</b></font>: {msg}')
				if m != c["msgs"][-1]:
					f.write("<br/>")
				
			f.write("</p>\n")

		# finish up
		f.write("</body></SAMI>")


def chat_to_ttml(msgs: List[ChatMessage], path: str, vid_duration:int, msg_duration:int):
	# DOES NOT CURRENTLY WORK
	# get chat with message in bounds
	chat_lists = chat_to_listwithbounds(msgs, vid_duration, msg_duration)

	with open(path, "w") as f:
		# write preamble stuffs
		f.write('<?xml version="1.0" encoding="utf-8"?>\n')
		f.write('<tt xmlns="http://www.w3.org/ns/ttml"><head>\n')
		f.write('<styling xmlns:tts="http://www.w3.org/ns/ttml#styling">\n')
		f.write('<style xml:id="s" tts:textAlign="left" tts:color="#ffffff"/>\n')
		f.write('</styling></head><body>\n')

		count = 0
		for c in chat_lists:
			f.write(f'<p xml:id="a{count}" style="s" begin="{c["begin"]}.0s" end="{c["end"]}.0s">')
			count += 1
			for m in c["msgs"]:
				msg = m["msg"].replace("<", "&lt;").replace(">", "&gt;").replace("&", "&amp;")
				f.write(f'<font color="#{m["clr"]}"><b>{m["usr"]}</b></font>: {msg}')
				if m != c["msgs"][-1]:
					f.write("<br/>")
				
			f.write("</p>\n")
		
		# finish up
		f.write("</body></tt>")


def timestring_as_seconds(time:str, default:int=0):
	if time == "EOF":
		return default
	
	s = time.split(":")
	
	seconds = int(s[-1]) if len(s) >= 1 else 0
	minutes = int(s[-2]) if len(s) >= 2 else 0
	hours = int(s[-3]) if len(s) >= 3 else 0

	# Hours, minutes, seconds
	return hours * 60 * 60 + minutes * 60 + seconds


def process_stage(conf: dict, stage: StageData, mode:str) -> Path:
	tempdir = Path(conf["temp_dir"])
	msg_duration = int(conf["chat_msg_time"])

	cprint(f"#rLoading all chat messages for `#fM{stage.id}#r`.", end="")
	total_offset = 0
	chat_list = []
	for slc in stage.slices:
		# load up each stagedata's meta to see if chat exists
		metapath = slc.filepath[:-4] + ".meta"
		meta = None
		with open(metapath, "r") as f:
			meta = json.load(f)
		
		has_chat = meta.get("has_chat", False)
		duration = meta.get("length", 0)
		chat_path = slc.filepath[:-4] + ".chat"
		# start and end times as seconds
		start_sec = timestring_as_seconds(slc.ss)
		end_sec = timestring_as_seconds(slc.to, duration)
		
		# keep each list of chat separate, compare timestamps to offsets to make
		# sure theyre inbetween the slices
		if has_chat:
			msg_list = [m for m in logfile_to_chat(chat_path) if start_sec <= m.offset < end_sec]
			if total_offset != 0:
				for m in msg_list:
					m.offset = m.offset - start_sec + total_offset
			
			chat_list += msg_list
		
		# now take each stage's duration and apply it to the next chat list's
		total_offset += end_sec - start_sec

	# determine how to export, then return the resultant path
	if mode != "upload" and mode != "export":
		util.exit_prog(94, f"Cannot export chat with export mode {mode}")
	
	export_type = conf["chat_"+mode]

	if len(chat_list) == 0:
		cprint(f" No chat found in `#fY{export_type}#r` stage. Skipping...")
		return False

	cprint(f" Exporting as format `#fY{export_type}#r`.")

	returnpath = None
	if export_type == "raw":
		# chat to logfile time
		returnpath = tempdir / (stage.id + ".chat")
		chat_to_logfile(chat_list, str(returnpath))
	elif export_type == "RealText":
		# load from archive, parse and write to temp
		returnpath = tempdir / (stage.id + ".rt")
		chat_to_realtext(chat_list, str(returnpath), total_offset, msg_duration)
	elif export_type == "SAMI":
		# load from archive, parse and write to temp
		returnpath = tempdir / (stage.id + ".sami")
		chat_to_sami(chat_list, str(returnpath), total_offset, msg_duration)
	elif export_type == "TTML":
		# load from archive, parse and write to temp
		returnpath = tempdir / (stage.id + ".ttml")
		chat_to_ttml(chat_list, str(returnpath), total_offset, msg_duration)

	return returnpath




from . import util, __project__, __version__
from .printer import colorize

import argparse
from pathlib import Path
from importlib import import_module
from requests.exceptions import ConnectionError


# Default path
vodbotdir = util.vodbotdir


def deffered_main():
	# Catch KeyboardInterrupts or connection failures, and report them cleanly
	try:
		main()
	except ConnectionError:
		util.exit_prog(-2, "Failed to connect to Twitch.")
	except KeyboardInterrupt:
		util.exit_prog(-1, "Interrupted by user.")


def main():
	titletext = colorize('#r#fM#l* {} {} (c) 2020-21 Logan "NotQuiteApex" Hickok-Dickson *#r')
	titletext = titletext.format(__project__, __version__)
	
	print(titletext)

	# Process arguments
	parser = argparse.ArgumentParser(epilog=titletext,
		description="Downloads and processes VODs and clips from Twitch.tv channels.")
	parser.add_argument("-v","--version", action="version", version=titletext)
	parser.add_argument("-c","--config", type=Path, dest="config",
		help="location of the Twitch config file to use", default=util.DEFAULT_CONF_PATH)

	# Subparsers for different commands
	subparsers = parser.add_subparsers(title="command", dest="cmd", help="command to run.")

	# `vodbot init`
	initparse = subparsers.add_parser("init", epilog=titletext,
		description="Runs the setup process for VodBot")
	initparse.add_argument("--output", help="path to save the config to; defaults to ~/.vodbot/conf.json",
		type=Path, required=False, default=util.DEFAULT_CONF_PATH, dest="output", metavar="PATH")
	initparse.add_argument("--default", action="store_true",
		help="argument switch to write default config; overrides all subsequent options")
	initparse.add_argument("--channels", type=str, default=[], nargs="+",
		help="twitch.tv channel login to pull VODs/Clips from", metavar="CHL")
	initparse.add_argument("--timezone", type=str, required=False, default=None,
		help="timezone for datetime referencing, as a UTC string")
	initparse.add_argument("--vod-dir", help="absolute path to store VOD data",
		type=Path, required=False, default=None, dest="voddir", metavar="PATH")
	initparse.add_argument("--clip-dir", help="absolute path to store Clip data",
		type=Path, required=False, default=None, dest="clipdir", metavar="PATH")
	initparse.add_argument("--temp-dir", help="absolute path to store temporary data",
		type=Path, required=False, default=None, dest="tempdir", metavar="PATH")
	initparse.add_argument("--stage-dir", help="absolute path to store staged data",
		type=Path, required=False, default=None, dest="stagedir", metavar="PATH")

	# `vodbot pull <vods/clips/both> [channel ...]`
	download = subparsers.add_parser("download", aliases=["pull", "download"],
		epilog=titletext, description="Downloads VODs and/or clips.")
	download.add_argument("type", type=str, default="both", nargs="?",
		help='content type flag, can be "vods", "clips", or "both"; defaults to "both"')
	download.add_argument("channels", metavar="channel", type=str, default=[], nargs="*",
		help="twitch.tv channel name to pull VODs from; optional, defaults to config")

	# `vodbot stage`
	stager = subparsers.add_parser("stage", epilog=titletext,
		description="Stages sections of video to upload or export",)
	stager_subparser = stager.add_subparsers(title="action", dest="action",
		description='action for staging the video')

	# `vodbot stage new \
	# `[--title "Apex - BBT"] [--desc "PogChamp {streamer}\n{link}"] \`
	# `<some_id> [--ss "0:0:0"] [--to "0:59:59"] \`
	# `<other_id> [--ss "0:20:0"] [--to "2:59:06"] \`
	# `<other_id> [--ss "3:20:0"] [--to "4:59:06"] \`
	# `<also_id> [--ss "0:40:0"] [--to "6:59:59"]`
	stager_add = stager_subparser.add_parser("new", epilog=titletext,
		description="creates a new stage for videos and clips to be mixed")
	stager_add.add_argument("id", help="id of the VOD or Clip to stage",
		type=str, nargs="+")
	stager_add.add_argument("--title", help="title of finished video",
		type=str, required=False, default="")
	stager_add.add_argument("--desc", help="description of finished video",
		type=str, required=False, default="")
	stager_add.add_argument("--ss", help="start time of video",
		type=str, required=False, default=[], nargs="?", action="append")
	stager_add.add_argument("--to", help="end time of video",
		type=str, required=False, default=[], nargs="?", action="append")

	# `vodbot stage rm <id>`
	stager_rm = stager_subparser.add_parser("rm", epilog=titletext,
		description="removes a VOD or Clip from the staging area")
	stager_rm.add_argument("id", type=str, help="id of the staged video data")

	# `vodbot stage list [id]`
	stager_list = stager_subparser.add_parser("list", epilog=titletext,
		description="lists info on staging area or staged items")
	stager_list.add_argument("id", nargs="?", type=str, help="id of the staged video data", default=None)

	# `vodbot upload <id/all>`
	upload = subparsers.add_parser("upload", aliases=["upload", "push"],
		epilog=titletext, description="Uploads stage(s) to YouTube.")
	upload.add_argument("id", type=str,
		help='id of the staged video data, "all" to upload all stages, or "logout" to remove existing YouTube credentials')
	
	# `vodbot export <id/all>`
	export = subparsers.add_parser("export", aliases=["export", "slice"],
		epilog=titletext, description="Uploads stage(s) to YouTube.")
	export.add_argument("id", type=str, help="id of the staged video data, or `all` for all stages")
	export.add_argument("path", type=Path, help="directory to export the video(s) to")

	# `vodbot upload <id/all>`
	info = subparsers.add_parser("info", epilog=titletext,
		description="Prints out info on the Channel, Clip, or VOD given.")
	info.add_argument("id", type=str, help="id/url of the Channel, Clip, or VOD")
	
	args = parser.parse_args()

	# Handle commands
	if args.cmd == "init":
		init = import_module(".commands.init", "vodbot")
		init.run(args)
	elif args.cmd == "pull" or args.cmd == "download":
		pull = import_module(".commands.pull", "vodbot")
		pull.run(args)
	elif args.cmd == "stage":
		stage = import_module(".commands.stage", "vodbot")
		stage.run(args)
	elif args.cmd == "upload" or args.cmd == "push":
		upload = import_module(".commands.upload", "vodbot")
		upload.run(args)
	elif args.cmd == "export" or args.cmd == "slice":
		export = import_module(".commands.export", "vodbot")
		export.run(args)
	elif args.cmd == "info":
		info = import_module(".commands.info", "vodbot")
		info.run(args)


if __name__ == "__main__":
	deffered_main()



# Based on https://github.com/ihabunek/twitch-dl/blob/master/twitchdl/download.py
# Modified to fit the project a bit better, originally licensed under GPLv3.
# Modifications include shortening some functions, removing redundant bits,
# and changing certain printouts to be more colorful.

from vodbot.printer import cprint

import os
import requests

from datetime import datetime
from collections import OrderedDict
from concurrent.futures import ThreadPoolExecutor, as_completed, wait
from functools import partial
from requests.exceptions import RequestException

CHUNK_SIZE = 1024
CONNECT_TIMEOUT = 5
RETRY_COUNT = 5

class DownloadFailed(Exception):
	pass

class DownloadCancelled(Exception):
	pass


def _download(url, path):
	tmp_path = path + ".tmp"
	response = requests.get(url, stream=True, timeout=CONNECT_TIMEOUT)
	size = 0
	with open(tmp_path, 'wb') as target:
		for chunk in response.iter_content(chunk_size=CHUNK_SIZE):
			target.write(chunk)
			size += len(chunk)

	os.rename(tmp_path, path)
	return size


def download_file(url, path, retries=RETRY_COUNT):
	if os.path.exists(path):
		return os.path.getsize(path), True

	for _ in range(retries):
		try:
			return _download(url, path), False
		except RequestException:
			pass

	raise DownloadFailed("sadge")

def format_duration(total_seconds):
	total_seconds = int(total_seconds)
	hours = total_seconds // 3600
	remainder = total_seconds % 3600
	minutes = remainder // 60
	seconds = total_seconds % 60

	if hours:
		return f"{hours} h {minutes} min"
	elif minutes:
		return f"{minutes} min {seconds} sec"
	else:
		return f"{seconds} sec"

def format_size(bytes_, digits=1):
	units = ["B", "kB", "MB"]
	for x in range(3):
		if bytes_ < 1024:
			if digits > 0:
				return "{{:.{}f}}{}".format(digits, units[x]).format(bytes_)
			else:
				return "{{:d}}{}".format(units[x]).format(bytes_)
		bytes_ /= 1024

	if digits > 0:
		return "{{:.{}f}}{}".format(digits, "GB").format(bytes_)
	else:
		return "{{:d}}{}".format("GB").format(bytes_)

def _print_progress(video_id, futures):
	downloaded_count = 0
	downloaded_size = 0
	existing_size = 0
	max_msg_size = 0
	start_time = datetime.now()
	total_count = len(futures)

	try:
		for future in as_completed(futures):
			(size, existed) = future.result()
			downloaded_count += 1
			downloaded_size += size
			if existed:
				existing_size += size

			percentage = 100 * downloaded_count // total_count
			est_total_size = int(total_count * downloaded_size / downloaded_count)
			duration = (datetime.now() - start_time).seconds
			speed = (downloaded_size-existing_size) // duration if duration else 0
			remaining = (total_count - downloaded_count) * duration / downloaded_count

			msg = " ".join([
				f"#fM#lVOD#r `#fM{video_id}#r` pt#fC{downloaded_count}#r/#fB#l{total_count}#r,",
				f"#fC{format_size(downloaded_size)}#r/#fB#l~{format_size(est_total_size)}#r #d({percentage}%)#r;",
				f"at #fY~{format_size(speed)}/s#r;" if speed > 0 else "",
				f"#fG~{format_duration(remaining)}#r left" if speed > 0 else "",
			])

			max_msg_size = max(len(msg), max_msg_size)
			cprint("\r" + msg.ljust(max_msg_size), end="")
	except KeyboardInterrupt:
		done, not_done = wait(futures, timeout=0)
		for future in not_done:
			future.cancel()
		wait(not_done, timeout=None)
		raise DownloadCancelled()


def download_files(video_id, base_url, target_dir, vod_paths, max_workers):
	urls = [base_url + path for path in vod_paths]
	targets = [str(target_dir / f"{k}.ts") for k, _ in enumerate(vod_paths)]
	partials = (partial(download_file, url, path) for url, path in zip(urls, targets))

	with ThreadPoolExecutor(max_workers=max_workers) as executor:
		futures = [executor.submit(fn) for fn in partials]
		_print_progress(video_id, futures)

	return OrderedDict(zip(vod_paths, targets))



#!/usr/bin/env python3
from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setup(
	name="vodbot",
	version="0.9.9",
    
	author="Logan \"NotQuiteApex\" Hickok-Dickson",
	author_email="self@notquiteapex.net",
    
	description="Twitch VOD and Clip manager",
    long_description=long_description,
	long_description_content_type="text/markdown",
    
	url="https://github.com/NotQuiteApex/VodBot",
    project_urls={
        "Bug Tracker": "https://github.com/NotQuiteApex/VodBot/issues",
    },
    
	keywords="twitch vod video download",
	license="zlib/libpng",
    
	packages=find_packages(include=["vodbot", "vodbot.*"]),
    
	python_requires='>=3.6',
	install_requires=[
		"google-api-python-client>=2.0",
		"google-auth-httplib2>=0.1.0",
		"google-auth-oauthlib>=0.4.4",
		"requests>=2.20",
		"m3u8>=0.8",
	],
	entry_points={
		'console_scripts': ['vodbot=vodbot.__main__:deffered_main'],
	}
)



# Staging, where videos get staged and set up with metadata to upload

import vodbot.util as util
from vodbot.printer import cprint, colorize

import re
import json
import string
import random
import datetime
from datetime import datetime, timezone, timedelta
from pathlib import Path
from os import remove as os_remove, listdir as os_listdir
from os.path import isfile, isdir
from typing import List


# Default path
vodbotdir = util.vodbotdir

DISALLOWED_CHARACTERS = [chr(x) for x in range(32)] # just in case...


class VideoSlice():
	def __init__(self, video_id: str, ss: str, to: str, filepath: str):
		self.video_id = video_id
		self.ss = ss
		self.to = to
		self.filepath = filepath
	
	def get_as_dict(self):
		return {"id":self.video_id, "ss":self.ss, "to":self.to, "path":str(self.filepath)}


class StageData():
	def __init__(self, streamers: List[str], title: str, desc: str, slices: List[VideoSlice], datestring: str, cid=None):
		for x in DISALLOWED_CHARACTERS:
			title = title.replace(x, "_")
		self.title = title
		self.desc = desc
		self.streamers = streamers
		self.datestring = datestring
		self.slices = slices

		if cid is None:
			self.gen_new_id()
		else:
			self.id = cid
	
	def __repr__(self):
		return f"StageData(\"{self.title}\", {self.id})"
	
	def write_stage(self, filename):
		with open(filename, "w") as f:
			jsondump = {
				"streamers": self.streamers,
				"title": self.title,
				"desc": self.desc,
				"datestring": self.datestring,
				"id": self.id,
				"slices": []
			}

			for vid in self.slices:
				jsondump["slices"] += [vid.get_as_dict()]

			json.dump(jsondump, f)
	
	def gen_new_id(self):
		self.id = ""
		for _ in range(4):
			self.id += random.choice(string.ascii_lowercase + string.digits)
	
	@staticmethod
	def load_from_json(data: dict) -> 'StageData':
		slices = [VideoSlice(v["id"], v["ss"], v["to"], v["path"]) for v in data["slices"]]
		streamers = data["streamers"]
		title = data["title"]
		desc = data["desc"]
		datestr = data["datestring"]
		_id = data["id"]

		new_data = StageData(streamers=streamers, title=title, desc=desc, datestring=datestr, slices=slices)
		new_data.id = _id

		return new_data
	
	@staticmethod
	def load_from_id(stagedir: Path, sid: str) -> 'StageData':
		jsonread = None
		try:
			with open(str(stagedir / (sid+".stage"))) as f:
				jsonread = json.load(f)
		except FileNotFoundError:
			util.exit_prog(46, f'Could not find stage "{sid}". (FileNotFound)')
		except KeyError:
			util.exit_prog(46, f'Could not parse stage "{sid}" as JSON. Is this file corrupted?')
		
		return StageData.load_from_json(jsonread)
	
	@staticmethod
	def load_all_stages(stagedir: Path) -> List['StageData']:
		stages = []
		for d in os_listdir(str(stagedir)):
			if isfile(str(stagedir / d)) and d[-5:] == "stage":
				stages.append(StageData.load_from_id(stagedir, d[:-6]))
		
		return stages


class CouldntFindVideo(Exception):
	pass


def create_format_dict(conf, streamers, utcdate=None, truedate=None):
	thistz = None
	datestring = None
	if truedate == None:
		try:
			# https://stackoverflow.com/a/37097784/13977827
			sign, hours, minutes = re.match('([+\-]?)(\d{2})(\d{2})', conf['stage_timezone']).groups()
			sign = -1 if sign == '-' else 1
			hours, minutes = int(hours), int(minutes)

			thistz = timezone(sign * timedelta(hours=hours, minutes=minutes))
		except:
			util.exit_prog(73, f"Unknown timezone {conf['stage_timezone']}")
		date = datetime.strptime(utcdate, "%Y-%m-%dT%H:%M:%SZ").replace(tzinfo=timezone.utc)
		datestring = date.astimezone(thistz).strftime("%Y/%m/%d")
	else:
		datestring = truedate
	
	formatdict = {
		"date": datestring,
		"link": f"https://twitch.tv/{streamers[0]}",
		"streamer": streamers[0],
		"links": " ".join([f"https://twitch.tv/{s}" for s in streamers]),
		"streamers": streamers,
	}

	# first pass format
	for x in range(2):
		for item, string in conf["stage_format"].items():
			try:
				string = string.format(**formatdict)
				formatdict[item] = string
			except KeyError as err:
				# ignore errors on first pass
				if x == 1:
					util.exit_prog(81, f"Format failed: {err}")

	return formatdict, datestring


def find_video_by_id(vid_id, VODS_DIR, CLIPS_DIR):
	"""
	Check where a video is by its ID

	:param vid_id: ID of the video.
	:returns: A tuple containing the path to the video file and the meta file.
	"""
	
	VODS_DIR_PATH = Path(VODS_DIR)
	CLIPS_DIR_PATH = Path(CLIPS_DIR)
	vod_dirs = [d for d in os_listdir(VODS_DIR) if isdir(str(VODS_DIR_PATH / d))]
	clip_dirs = [d for d in os_listdir(CLIPS_DIR) if isdir(str(CLIPS_DIR_PATH / d))]

	filename = None
	
	directories = [(vod_dirs, VODS_DIR_PATH), (clip_dirs, CLIPS_DIR_PATH)]

	for dir_t in directories:
		for channel in dir_t[0]:
			folder = dir_t[1] / Path(channel)
			metas = [m[:-5] for m in os_listdir(str(folder)) if isfile(str(folder / Path(m))) and m[-4:]=="meta"]
			metas = [m for m in metas if vid_id in m] # multiple types of id's exist, so we have to soft match
			if len(metas) > 0:
				vid_id = metas[0] # use first result of matching
				metajson = None
				try:
					with open(str(folder / (vid_id+".meta"))) as f:
						metajson = json.load(f)
					filename = folder / f"{metajson['created_at']}_{metajson['id']}.mkv".replace(":", ";")
					return (filename, metajson, "VOD" if dir_t[1] == VODS_DIR_PATH else "Clip")
				except FileNotFoundError:
					pass
				except ValueError:
					pass
	
	raise CouldntFindVideo()


def check_stage_id(stage_id, STAGE_DIR):
	STAGE_DIR_PATH = Path(STAGE_DIR)

	stages = [m[:-6] for m in os_listdir(STAGE_DIR) if isfile(str(STAGE_DIR_PATH / Path(m))) and m[-5:]=="stage"]

	return stage_id in stages


def check_time(prefix, resp, default=None):
	output = resp
	checkedonce = False

	while True:
		if checkedonce or not output:
			a = 'ss' if prefix=='Start' else 'to'
			t = '0:0:0' if prefix=='Start' else 'EOF'
			f = f"#fW#l{prefix} time of the Video#r #d(--{a}, default {t})#r: "
			output = input(colorize(f))
		checkedonce = True

		if output == "":
			if prefix == "Start":
				return default if default != None else "0:0:0"
			elif prefix == "End":
				return default if default != None else "EOF"

		intime = output.split(":")
		timelist = []
		seconds = None
		minutes = None
		hours = None

		if len(intime) > 3:
			cprint(f"#fR{prefix} time: Time cannot have more than 3 units.#r")
			continue
		
		if len(intime) >= 1:
			seconds = intime[-1]
			try:
				seconds = int(seconds)
			except ValueError:
				cprint(f"#fR{prefix} time: Seconds does not appear to be a number.#r")
				continue
			if seconds > 59 or seconds < 0:
				cprint(f"#fR{prefix} time: Seconds must be in the range of 0 to 59.#r")
				continue
			timelist.insert(0, str(seconds))
		
		if len(intime) >= 2:
			minutes = intime[-2]
			try:
				minutes = int(minutes)
			except ValueError:
				cprint(f"#fR{prefix} time: Minutes does not appear to be a number.#r")
				continue
			if minutes > 59 or minutes < 0:
				cprint(f"#fR{prefix} time: Minutes must be in the range of 0 to 59.#r")
				continue
			timelist.insert(0, str(minutes))
		else:
			timelist.insert(0, "0")
		
		if len(intime) == 3:
			hours = intime[-3]
			try:
				hours = int(hours)
			except ValueError:
				cprint(f"#fR{prefix} time: Hours does not appear to be a number.#r")
				continue
			timelist.insert(0, str(hours))
		else:
			timelist.insert(0, "0")
		
		output = ":".join(timelist)
		break

	return output


def check_streamers(default=None) -> List[str]:
	streamers = ""
	while not streamers:
		streamers = input(colorize(f"#fW#lWho was in the VOD#r #d(default `{', '.join(default)}`, csv)#r: "))

		if streamers == "":
			streamers = default
		else:
			streamers = streamers.replace(" ", "").split(",")
			for streamer in streamers:
				if len(streamer) == 0:
					cprint("#l#fRMissing streamer name!#r")
					streamers = ""
					break
	
	return streamers


RESERVED_NAMES = [
	"CON", "PRN", "AUX", "NUL",
	"COM1", "COM2", "COM3", "COM4", "COM5", "COM6", "COM7", "COM8", "COM9",
	"LPT1", "LPT2", "LPT3", "LPT4", "LPT5", "LPT6", "LPT7", "LPT8", "LPT9",
]
def check_title(default=None):
	title = default
	if not title:
		title = ""
		while title == "":
			title = input(colorize("#fW#lTitle of the Video#r #d(--title)#r: "))
			# blank title
			if title == "":
				cprint("#fRTitle cannot be blank.#r")
			# reserved names
			if title in RESERVED_NAMES:
				cprint("#fRTitle cannot be a reserved name.#r")
				title = ""
	
	return title


def check_description(formatdict, inputdefault=None):
	desc = ""

	if inputdefault:
		try:
			inputdefault = inputdefault.format(**formatdict).replace("\\n", "\n")
			desc = inputdefault
		except KeyError as err:
			cprint(f"#fRDescription format error from default: {err}.#r")
			desc = ""
	
	while desc == "":
		desc = input(colorize("#fW#lDescription of Video#r #d(--desc)#r: "))
		if desc == "":
			cprint("#fRDescription cannot be blank.#r")
			continue

		# Format the description
		try:
			desc = desc.format(**formatdict).replace("\\n", "\n")
		except KeyError as err:
			cprint(f"#fRDescription format error: {err}.#r")
			desc = ""
	
	return desc


def _new(args, conf, stagedir):
	VODS_DIR = conf["vod_dir"]
	CLIPS_DIR = conf["clip_dir"]
	STAGE_DIR = conf["stage_dir"]
	stagedir = Path(STAGE_DIR)

	# find the videos by their ids to confirm they exist
	videos = []
	for video in args.id:
		try:
			(filename, metadata, videotype) = find_video_by_id(video, VODS_DIR, CLIPS_DIR)
			videos += [{"id":video, "file":filename, "meta":metadata, "type":videotype}]
		except CouldntFindVideo:
			util.exit_prog(13, f'Could not find video with ID "{args.id}"')
	
	# Get what streamers were involved (usernames), always asked
	default_streamers = []
	for f in videos:
		if f["meta"]["user_login"] not in default_streamers:
			default_streamers.append(f["meta"]["user_login"])
	args.streamers = check_streamers(default=default_streamers)

	# get title
	if not args.title:
		args.title = check_title(default=None)

	# get description
	formatdict, datestring = create_format_dict(conf, args.streamers, utcdate=metadata["created_at"])
	args.desc = check_description(formatdict, inputdefault=args.desc)

	# get timestamps for each video through input
	for x in range(len(videos)):
		# skip times we dont need because we already have them
		if x < len(args.ss):
			continue
		
		vid = videos[x]
		# grab times for this specific stream
		cprint(f"#dTimestamps for `#r#fM{vid['meta']['title']}#r` #d({vid['meta']['id']})#r")
		args.ss += [check_time("Start", args.ss[x] if x < len(args.ss) else None)]
		args.to += [check_time("End", args.to[x] if x < len(args.to) else None)]

	# make slice objects
	slices = []
	for x in range(len(videos)):
		vid = videos[x]
		vidslice = VideoSlice(video_id=vid["id"], ss=args.ss[x], to=args.to[x], filepath=vid["file"])
		slices += [vidslice]

	# make stage object
	stage = StageData(streamers=args.streamers, title=args.title, desc=args.desc, datestring=datestring, slices=slices)
	# Check that new "id" does not collide
	while check_stage_id(stage.id, STAGE_DIR):
		stage.gen_new_id()

	# shorter file name
	#shortfile = stage.filename.replace(VODS_DIR, "$vods").replace(CLIPS_DIR, "$clips")

	print()
	cprint(f"#r`#fC{stage.title}#r` #fM{' '.join(stage.streamers)}#r #d({stage.id})#r")
	cprint(f"#d'''#fG{stage.desc}#r#d'''#r")
	for vid in stage.slices:
		cprint(f"#fM{vid.video_id}#r > #fY{vid.ss}#r - #fY{vid.to}#r")
	
	# write stage
	stagename = str(stagedir / (stage.id + ".stage"))
	stage.write_stage(stagename)

	# Done!


def _list(args, conf, stagedir):
	VODS_DIR = conf["vod_dir"]
	CLIPS_DIR = conf["clip_dir"]
	STAGE_DIR = conf["stage_dir"]
	stagedir = Path(STAGE_DIR)
	
	if args.id == None:
		stages = StageData.load_all_stages(stagedir)

		for s in stages:
			cprint(f'#r#fY#l{s.id}#r -- `#fC{s.title}#r` ', end="")
			cprint(f'(#fM{" ".join([d.video_id for d in s.slices])}#r) -- ', end="")
			cprint(f'#l#fM{", ".join(s.streamers)}#r')
		
		if len(stages) == 0:
			cprint("#fBNothing staged right now.#r")
	else:
		stage = StageData.load_from_id(stagedir, args.id)

		cprint(f"#r`#fC{stage.title}#r` #fM{' '.join(stage.streamers)}#r #d({stage.id})#r")
		cprint(f"#d'''#fG{stage.desc}#r#d'''#r")
		for vid in stage.slices:
			cprint(f"#fM{vid['id']}#r > #fY{vid['ss']}#r - #fY{vid['to']}#r")


def run(args):
	conf = util.load_conf(args.config)
	
	util.make_dir(vodbotdir)
	stagedir = conf["stage_dir"]
	path_stage = Path(stagedir)
	util.make_dir(stagedir)

	if args.action == "new":
		_new(args, conf, stagedir)
	elif args.action == "rm":
		if not isfile(str(path_stage / (args.id + ".stage"))):
			util.exit_prog(45, f'Could not find stage "{args.id}".')
			
		try:
			os_remove(str(path_stage / (args.id + ".stage")))
			cprint(f'Stage "#fY#l{args.id}#r" has been #fRremoved#r.')
		except OSError as err:
			util.exit_prog(88, f'Stage "{args.id}" could not be removed due to an error. {err}')
	elif args.action == "list":
		_list(args, conf, stagedir)



from vodbot import util, twitch
from vodbot.itd import download as itd_dl, worker as itd_work
from vodbot.twitch import Channel, Vod, Clip
from vodbot.printer import cprint

from pathlib import Path
from os import listdir
from os.path import isfile


# Default path
vodbotdir = util.vodbotdir


def run(args):
	download_twitch_video(args)


def download_twitch_video(args):
	# Load the config and set up the access token
	cprint("#r#dLoading config...#r", end=" ", flush=True)
	conf = util.load_conf(args.config)
	CHANNEL_IDS = conf["twitch_channels"]
	VODS_DIR = conf["vod_dir"]
	CLIPS_DIR = conf["clip_dir"]
	TEMP_DIR = conf["temp_dir"]
	LOG_LEVEL = conf["ffmpeg_loglevel"]
	PULL_CHAT = conf["pull_chat_logs"]

	# If channel arguments are provided, override config
	if args.channels:
		CHANNEL_IDS = args.channels
	
	cprint("#r#dLoading channel data...#r", flush=True)
	channels = twitch.get_channels(CHANNEL_IDS)
	
	contentnoun = "video" # temp contentnoun until the rest is reworked
	
	# Setup directories for videos and temp
	util.make_dir(TEMP_DIR)
	voddir = Path(VODS_DIR)
	util.make_dir(voddir)
	clipdir = Path(CLIPS_DIR)
	util.make_dir(clipdir)

	# Get list of videos using channel object ID's from Twitch API
	videos = []
	totalvods = 0
	totalclips = 0

	channel_print = None
	if args.type == "both":
		channel_print = "Pulling #fM#lVOD#r & #fM#lClip#r list: #fY#l{}#r..."
	elif args.type == "vods":
		channel_print = "Pulling #fM#lVOD#r list: #fY#l{}#r..."
	elif args.type == "clips":
		channel_print = "Pulling #fM#lClip#r list: #fY#l{}#r..."

	for channel in channels:
		vods = None
		clips = None
		cprint(channel_print.format(channel.display_name), end=" ")

		# Grab list of VODs and check against existing VODs
		if args.type == "both" or args.type == "vods":
			folder = voddir / channel.login
			util.make_dir(folder)
			allvods = twitch.get_channel_vods(channel)
			vods = compare_existant_file(folder, allvods)
			totalvods += len(vods)

		# Grab list of Clips and check against existing Clips
		if args.type == "both" or args.type == "clips":
			folder = clipdir / channel.login
			util.make_dir(folder)
			allclips = twitch.get_channel_clips(channel)
			clips = compare_existant_file(folder, allclips)
			totalclips += len(clips)

		# Print content found and save it
		if args.type == "both":
			cprint(f"#fC#l{len(vods)} #fM#lVODSs#r & #fC#l{len(clips)} #fM#lClips#r")
			videos += vods
			videos += clips
		elif args.type == "vods":
			cprint(f"#fC#l{len(vods)} #fM#lVODSs#r")
			videos += vods
		elif args.type == "clips":
			cprint(f"#fC#l{len(clips)} #fM#lClips#r")
			videos += clips
			
	if args.type == "both":
		cprint(f"Total #fMVODs#r to download: #fC#l{totalvods}#r")
		cprint(f"Total #fMClips#r to download: #fC#l{totalclips}#r")
		cprint(f"Total #fM#lvideos#r to download: #fC#l{len(videos)}#r")
	elif args.type == "vods":
		cprint(f"Total #fMVODs#r to download: #fC#l{totalvods}#r")
	elif args.type == "clips":
		cprint(f"Total #fMClips#r to download: #fC#l{totalclips}#r")

	# Download all the videos we need.
	previouschannel = None
	for vod in videos:
		# Print if we're on to a new user.
		if previouschannel != vod.user_id:
			previouschannel = vod.user_id
			cprint(f"\nDownloading #fY#l{vod.user_name}#r's #fM#l{contentnoun}s#r...")

		# Generate path for video
		viddir = None
		contentnoun = None

		if isinstance(vod, Vod):
			viddir = voddir / vod.user_name.lower()
			contentnoun = "VOD"
		elif isinstance(vod, Clip):
			viddir = clipdir / vod.user_name.lower()
			contentnoun = "Clip"
		
		filepath = viddir / f"{vod.created_at}_{vod.id}".replace(":", ";")
		filename = str(filepath) + ".mkv"
		metaname = str(filepath) + ".meta"
		chatname = str(filepath) + ".chat"

		# Write video data and handle exceptions
		try:
			if isinstance(vod, Vod):
				# download chat
				if PULL_CHAT:
					itd_dl.dl_video_chat(vod, chatname)
					vod.has_chat = True
				# download video
				itd_dl.dl_video(vod, Path(TEMP_DIR), filename, 20, LOG_LEVEL)
				# write meta file
				vod.write_meta(metaname)
			elif isinstance(vod, Clip):
				# download clip
				itd_dl.dl_clip(vod, filename)
				# write meta file
				vod.write_meta(metaname)
		except itd_dl.JoiningFailed:
			cprint(f"#fR#lVOD `{vod.id}` joining failed! Preserving files...#r")
		except itd_work.DownloadFailed:
			cprint(f"Download failed! Skipping...#r")
		except itd_work.DownloadCancelled:
			cprint(f"\n#fR#l{contentnoun} download cancelled.#r")
			raise KeyboardInterrupt()
	
	cprint("\n#fM#l* All done, goodbye! *#r\n")


def compare_existant_file(path, allvods):
	# Check for existing videos by finding the meta files
	existingvods = [f[:-5] for f in listdir(str(path)) if isfile(str(path/f)) and f[-4:]=="meta"]
	# Compare vods, if they arent downloaded (meta is missing) then we need to queue them
	result = [
		vod 
			for vod in allvods
			# we use two methods of detection, one for the new format and one for the old
			# this is so old archives don't get overwritten unecessarily
			# TODO: remove this in a future update after deprecation
			if not any(
				f"{vod.created_at}_{vod.id}".replace(":", ";") == x
				or vod.id == x for x in existingvods
					for x in existingvods
			)
		]
	return result
